#!/bin/dash
# download and install latest Brave browser
# can be used to package existing "portable linux" tarball, pass filename as $1. 
# jamesbond 2024 (+ JakeSFR 2025 - adapted for Brave)
# MIT license

. gettext.sh
export TEXTDOMAIN=get_app

### config
APPTITLE="$(gettext 'Download Brave Browser')"
PKGNAME=brave
WORKDIR=/tmp/work/${PKGNAME}.$$
PKGFILE=${1}

### generic helpers
get_config() {
	if [ $(id -u) -ne 0 ]; then
		msg "$(gettext 'You must be root.')"
		exit 1
	fi
	LIBDIR=lib64
	ARCH=x86_64

	local DOWNLOADER="wget -qO -"
	local VER_URL="https://api.github.com/repos/brave/brave-browser/releases/latest"
	LATEST= DL_URL=
	
	if [ -z "$PKGFILE" ]; then
		DL_URL="$($DOWNLOADER ${VER_URL} | sed -n 's#.*browser_download_url.*\(https://.*linux-amd64.zip\)".*#\1#p')"
		LATEST="$(echo "$DL_URL" | sed -n 's#.*releases/download/v\(.*\)/brave.*#\1#p')"
		PKGFILE="${DL_URL##*/}"
	fi
}

# $1-text
msg() {
	[ -z "$DISPLAY" ] && echo "$1" ||
	Xdialog --title "$APPTITLE" --infobox "$1" 0 0 10000
}

# $1-text, returns true/false
yesno() {
	local p
	if [ -z "$DISPLAY" ]; then
		printf "%s\nType [yes] to continue: " "$1"
		read p
		case "$p" in
			[Yy][Ee][Ss]) return 0
		esac
		return 1
	else
		Xdialog --title "$APPTITLE" --yesno "$1" 0 0
	fi
}

# $1-text
splash() {
	if [ -z "$DISPLAY" ]; then
		echo "$1" 
	else
		if [ "$1" ]; then
			[ $SPID ] && kill $SPID
			Xdialog --title "$APPTITLE" --no-buttons --infobox "$1" 0 0 10000000 &
			SPID=$!
		else
			[ $SPID ] && kill $SPID
			SPID=""
		fi
	fi
}

# $1 - path to brave-browser zip archive
package_brave() {
	splash "$(eval echo $(gettext 'Making $PKGNAME package, the package will be stored in $HOME'))"
	mkdir pkg.$$; cd pkg.$$
	
	PKGVER=$(ls -d ../${PKGNAME}* | sed 's/.*brave-browser-//; s/-.*//')
	
	# extract
	mkdir "$PKGNAME-$PKGVER"; cd "$PKGNAME-$PKGVER"
	unzip "$1"
	cd ..

	# setup symlinks, etc
	mkdir -p opt usr/share/icons/hicolor/48x48/apps usr/bin usr/share/applications
	ln -s ${PKGNAME}* opt/$PKGNAME
	mv ${PKGNAME}* opt
	cp opt/$PKGNAME/product_logo_48.png usr/share/icons/hicolor/48x48/apps/brave.png
	   
	# run brave as spot
	cat << "EOF" >> usr/bin/brave-spot &&
#!/bin/dash
SPOT_HOME=$(awk -F: '$1=="spot" {print $6}' /etc/passwd)
case "$TZ" in
	:*) TZ="${TZ#:}"
	test -L "$TZ" && TZ=$(readlink "$TZ")
	TZ=${TZ#/usr/share/zoneinfo/}
esac
exec /usr/bin/run-as-spot /opt/brave/brave-browser --user-data-dir="$SPOT_HOME/.config/brave" "$@"
EOF

	cat << "EOF" >> usr/bin/brave &&
#!/bin/dash
case "$TZ" in
	:*) TZ="${TZ#:}"
	test -L "$TZ" && TZ=$(readlink "$TZ")
	TZ=${TZ#/usr/share/zoneinfo/}
esac
exec /opt/brave/brave-browser "$@"
EOF

	chmod +x usr/bin/brave-spot usr/bin/brave &&
	
	# desktop file
	cat << "EOF" > usr/share/applications/$PKGNAME.desktop 
[Desktop Entry]
Version=1.0
Name=Brave Browser
GenericName=Web Browser
Comment=Chromium based browser
Exec=brave-spot %U
Icon=brave.png
Terminal=false
Type=Application
MimeType=text/html;text/xml;application/xhtml+xml;application/vnd.mozilla.xul+xml;text/mml;x-scheme-handler/http;x-scheme-handler/https;
StartupNotify=false
X-MultipleArgs=false
X-Desktop-File-Install-Version=0.16
Categories=WebBrowser;
Encoding=UTF-8
EOF

	# doinst.sh & desc 
	mkdir install
cat << EOF > install/slack-desc
$PKGNAME: $PKGNAME $PKGVER (Web browser)  
$PKGNAME:  
$PKGNAME: Web browser based on Chromium that blocks ads and trackers by default.
$PKGNAME: 
$PKGNAME: https://brave.com/
$PKGNAME: 
$PKGNAME: 
$PKGNAME: 
$PKGNAME: 
$PKGNAME: 
$PKGNAME:  
EOF
	cat << EOF > install/doinst.sh
#!/bin/sh
chown -RH root ./opt/${PKGNAME}/
chmod 4755 ./opt/$PKGNAME/chrome-sandbox
EOF

	# make the pkg
	makepkg -c n -l n $HOME/$PKGNAME-$PKGVER-$ARCH-official.txz > /dev/null 2>&1
}

install_brave() {
	splash "$(eval echo $(gettext 'Installing Brave browser $PKGVER...'))"
	removepkg $PKGNAME
	installpkg $HOME/$PKGNAME-$PKGVER-$ARCH-official.txz
	splash
	msg "$(gettext 'Done.')"
}

download_gtk3() {
	if ! slapt-get --installed | grep -q '^gtk3.*x86_64'; then
		splash "$(gettext "Updating package list, please wait...")"
		slapt-get --update --no-prompt
		splash "$(gettext "Installing GTK+3, please wait...")"
		if ! slapt-get --install --no-prompt gtk3; then
			splash
			msg "$(gettext "Failed to install GTK+3.")"
			exit 1
		fi
	fi
}

# DL_URL
download_brave() {
	wget -q "${DL_URL}"
}



#### main
get_config
if [ -z "$PKGFILE" ]; then
	splash "Unable to find download links, quitting ..."
	exit 1
fi

if [ -n "$DL_URL" ]; then
! yesno "\
$(gettext 'Running this program will download, package, and install the latest Brave browser.')
$(eval_gettext 'Latest version is: $LATEST.')

$(gettext 'If you already have Brave installed it will be replaced.') 
$(eval_gettext 'The newly created package can be found in $HOME.') 
$(gettext 'You can can save this package for later or just delete it.') 
$(gettext 'Are you sure you want to continue?')
" && exit
fi

download_gtk3
mkdir -p $WORKDIR; cd $WORKDIR
if [ -n "$DL_URL" ]; then
	splash "$(eval_gettext 'Downloading Brave $LATEST, please wait ...')"
	if ! download_brave; then
		splash
		msg "$(gettext  'Failed to download.')"
		cd; rm -rf $WORKDIR
		exit 1
	fi
fi

package_brave $(readlink -f "$PKGFILE")
install_brave
cd; rm -rf $WORKDIR

